%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  A Global Optimal Solution to the Eco-Driving Problem
%  G.P. Padilla, S. Weiland, M.C.F. Donkers
%
%  Numerical Example A. Benchmark Problem for Electric Vehicles.
%  
%


clear all; close all; clc
par = problem_configA();
par = parametersA(par);
N=par.prb.N;


%Dynamical Model  Ad x = Bd; with x=[a;s;v] ==

as=zeros(N,N+1);
ss=full(spdiags([ones(N+1,1),-ones(N+1,1)],[0,1],N,N+1));
vs=full(spdiags(par.prb.Ts*ones(N+1,1),0,N,N+1));
av=vs;
sv=as;
vv=ss;
Ad=[as ss vs; av sv vv];
bd=zeros(2*N,1);


%Contraints ===

    % Intial and final limits
     Al=[ zeros(1,N+1), 1, zeros(1,N), zeros(1,N+1) ;
          zeros(1,N+1), zeros(1,N), 1, zeros(1,N+1) ;
          zeros(1,N+1), zeros(1,N+1), 1, zeros(1,N) ;
          zeros(1,N+1), zeros(1,N+1), zeros(1,N), 1 ];
     bl=[ par.veh.s0; par.veh.sN; par.veh.v0; par.veh.vN];
     
    % Bounds
      lb=[];
      ub=[];
    Aineq=[eye(N+1) zeros(N+1) zeros(N+1);
           -eye(N+1) zeros(N+1) zeros(N+1);
           zeros(N+1) zeros(N+1) eye(N+1);
           zeros(N+1) zeros(N+1) -eye(N+1)];
    bineq= [ par.veh.aMax*ones(N+1,1);
            -par.veh.aMin*ones(N+1,1)
             par.veh.vMax*ones(N+1,1);
            -par.veh.vMin*ones(N+1,1);];



%SQP === 
% Initial guess
a = 0*ones(N+1,1);
s = 0*ones(N+1,1);
v = 0*ones(N+1,1);
x=[a;s;v];

% Road profile (from Petit's paper)
gamma0= par.pet.p0+par.pet.p1*s+par.pet.p2*s.^2+par.pet.p3*s.^3;
gamma0_prima=par.pet.p1+2*par.pet.p2*s+3*par.pet.p3*s.^2;
gamma0_prima2= 2*par.pet.p2+6*par.pet.p3*s;


P=0; %Cost
stop=0; %Stop flag
j=1; %Iterations
options = optimoptions('quadprog','Display','off');

tic
while ~stop
   
   
    %Hessian
    h11 =diag(par.em.beta2*2*par.veh.m^2*ones(N+1,1));
    h12=diag(zeros(N+1,1));
    h13=diag(zeros(N+1,1));
    h22=diag(1e-4*ones(N+1,1));
    h23=diag(zeros(N+1,1));
    h33 = diag(  max( 1e-5 , 2*par.em.beta0 + par.em.beta2*(12*par.veh.Cd^2*v.^2 + ...
            4*par.veh.Cd*par.veh.Cr + 4*gamma0*par.veh.g*par.veh.m*par.veh.Cd) + 6*par.veh.Cd*par.em.beta1*v ) );
    
    
    %Gradient
    f1 = (par.em.beta2*(2*par.veh.m^2*a + 2*gamma0*par.veh.g*par.veh.m^2));
    f2 = (2*par.em.beta2*par.veh.g*par.veh.m*gamma0_prima.*(a*par.veh.m + ...
            par.veh.Cd*v.^2 + par.veh.Cr + par.veh.g*par.veh.m*gamma0));%zeros(N,1);
    f3 = (par.em.beta1*(3*par.veh.Cd*v.^2 + par.veh.Cr) + 2*par.em.beta0*v +...
            par.em.beta2*(4*par.veh.Cd^2*v.^3 + 4*par.veh.Cd*par.veh.Cr.*v + 4*par.veh.g*par.veh.m*par.veh.Cd*v));
    
   
       
    H=[h11 h12 h13; 
       h12 h22 h23; 
       h13 h23 h33];
    f= [f1; f2; f3];
    
        
      %Optimization==
      aux=x;
      
      x = quadprog(1*H,f'-x'*1*H,Aineq,bineq,[Ad;Al],[bd;bl],lb,ub,[],options);

      %Updating variables==
       P_old=P;

       a=x(1:N+1);
       s=x(N+1+1:2*(N+1));
       v=x(2*(N+1)+1:end);
 
       gamma0= par.pet.p0+par.pet.p1*s+par.pet.p2*s.^2+par.pet.p3*s.^3;
       gamma0_prima=par.pet.p1+2*par.pet.p2*s+3*par.pet.p3*s.^2;
       gamma0_prima2= 2*par.pet.p2+6*par.pet.p3*s;
      

       u = a*par.veh.m+par.veh.Cd*v.^2+par.veh.Cr+par.veh.m*par.veh.g*gamma0;
       P=2*0.5*par.prb.Ts*sum(par.em.beta0*v.^2+par.em.beta1*u.*v+par.em.beta2*u.^2);
      

      %Conditions==
      
      if norm(aux-x) < 1e-6 && norm(P-P_old)<1e-6
          stop=1;
      end
      j=j+1;
end
toc


fprintf('Iterations: %d\n',j);

%Updating final solution
vf = [v(:,end); par.veh.vN];
sf = [s(:,end); par.veh.sN];
vdot = (vf(2:end,end)-vf(1:end-1,end))/par.prb.Ts;
u = vdot*par.veh.m+par.veh.Cd*vf(1:end-1,end).^2+par.veh.Cr+par.veh.m*par.veh.g*gamma0(:,end);
cost = par.prb.Ts*sum(par.em.beta2*u.^2+par.em.beta1*u.*vf(1:end-1)+par.em.beta0*vf(1:end-1).^2);
fprintf('Final cost: %5e\n',cost);

    

%% Plots on Time domain
t=[0:par.prb.Ts:par.prb.Tmax]';
vc=v(:,end);
ac=diff([vc;par.veh.vN]/par.prb.Ts);
sc=0;
for i=1:1:N
    sc(i+1)=sc(i)+vc(i)*par.prb.Ts;
end
gammac= [par.pet.p0+par.pet.p1*sc+par.pet.p2*sc.^2+par.pet.p3*sc.^3]';
uc=ac*par.veh.m+par.veh.Cd*vc.^2+par.veh.Cr+par.veh.m*par.veh.g*gammac;


figure
plot(t,sc,'LineWidth',2)
hold on
plot(t,vc,'LineWidth',2)
hold on
plot(t,gammac,'LineWidth',2)
hold on
plot(t,uc/10,'LineWidth',2)
xlabel('Time')
legend({'$s$','$v$','$\gamma$','$\frac{u}{10}$'},'interpreter','latex');
grid on

