%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  A Global Optimal Solution to the Eco-Driving Problem
%  G.P. Padilla, S. Weiland, M.C.F. Donkers
%
%  Numerical Example B. Hybrid Electric Heavy-Duty Vehicle.
%  
%

clear all; close all; clc
par = problem_configB();
par = parametersB(par);



N=par.prb.N;

%Dynamic Model ==

as=zeros(N,N+1);
ss=full(spdiags([ones(N+1,1),-ones(N+1,1)],[0,1],N,N+1));
vs=full(spdiags(par.prb.Ts*ones(N+1,1),0,N,N+1));
av=vs;
sv=as;
vv=ss;
Ad=[as ss vs; av sv vv];
bd=zeros(2*N,1);


%Contraints ===

    % Intial and final limits
     Al=[ zeros(1,N+1), 1, zeros(1,N), zeros(1,N+1) ;
          zeros(1,N+1), zeros(1,N), 1, zeros(1,N+1) ;
          zeros(1,N+1), zeros(1,N+1), 1, zeros(1,N) ;
          zeros(1,N+1), zeros(1,N+1), zeros(1,N), 1 ];
     bl=[ par.veh.s0; par.veh.sN; par.veh.v0; par.veh.vN];
     
    % Bounds
      lb=[];
      ub=[];
    Aineq=[eye(N+1) zeros(N+1) zeros(N+1);
           -eye(N+1) zeros(N+1) zeros(N+1);
           zeros(N+1) zeros(N+1) eye(N+1);
           zeros(N+1) zeros(N+1) -eye(N+1)];
    bineq= [ par.veh.aMax*ones(N+1,1);
            -par.veh.aMin*ones(N+1,1)
             par.veh.vMax*ones(N+1,1);
            -par.veh.vMin*ones(N+1,1);];



%SQP === 
% Initial guess
a = 0*ones(N+1,1);
s = 0*ones(N+1,1);
v = 0*ones(N+1,1);
x=[a;s;v];
% Road profile
gamma0= par.veh.beta*sin(par.veh.gamma1*s+par.veh.gamma0);
gamma0_prima= par.veh.beta.*cos(par.veh.gamma1*s+par.veh.gamma0)*par.veh.gamma1;
gamma0_prima2= -par.veh.beta.*cos(par.veh.gamma1*s+par.veh.gamma0)*par.veh.gamma1*par.veh.gamma1;

   
options = optimoptions('quadprog','Display','off');
P=0;
stop=0;
j=1;


P_min=0;
while ~stop
   
   
    %Hessian
    h11 = diag(par.em.beta2*2*par.veh.m^2*ones(N+1,1));
    h12=diag(zeros(N+1,1));
    h13=diag(zeros(N+1,1));
    h22=diag(1e-8*ones(N+1,1));
    h23=diag(zeros(N+1,1));
    h33 = (2*par.em.beta0 + par.em.beta2*(12*par.veh.Cd^2*v.^2 + ...
            4*par.veh.Cd*par.veh.Cr + 4*gamma0*par.veh.g*par.veh.m*par.veh.Cd) + 6*par.veh.Cd*par.em.beta1*v);
    h33=max(h33,1e-8*ones(N+1,1));
    h33=diag(h33);
    
    %Gradient
    f1 = (par.em.beta2*(2*par.veh.m^2*a + 2*gamma0*par.veh.g*par.veh.m^2));
    f2 = (2*par.em.beta2*par.veh.g*par.veh.m*gamma0_prima.*(a*par.veh.m + ...
            par.veh.Cd*v.^2 + par.veh.Cr + par.veh.g*par.veh.m*gamma0));%zeros(N,1);
    f3 = (par.em.beta1*(3*par.veh.Cd*v.^2 + par.veh.Cr) + 2*par.em.beta0*v +...
          par.em.beta2*(4*par.veh.Cd^2*v.^3 + 4*par.veh.Cd*par.veh.Cr.*v + 4*par.veh.g*par.veh.m*par.veh.Cd*v.*gamma0));
    
   
  
    H=[h11 h12 h13; 
       h12 h22 h23; 
       h13 h23 h33];
    f= [f1; f2; f3];
    
    

       
      %Optimization==
      aux=x;
      [x,fval] = quadprog(H,f'-x'*H,Aineq,bineq,[Ad;Al],[bd;bl],lb,ub,[],options);
      %[x,fval,exitflag,output,lambda]  = cplexqp(H,f'-x'*H,Aineq,bineq,[Ad;Al],[bd;bl],lb,ub);
  
      P_old=P;
      a=x(1:N+1);
      s=x(N+1+1:2*(N+1));
      v=x(2*(N+1)+1:end);
      
      gamma0= par.veh.beta*sin(par.veh.gamma1*s+par.veh.gamma0);
      gamma0_prima= par.veh.beta.*cos(par.veh.gamma1*s+par.veh.gamma0)*par.veh.gamma1;
      gamma0_prima2= -par.veh.beta.*cos(par.veh.gamma1*s+par.veh.gamma0)*par.veh.gamma1*par.veh.gamma1;

      u = a*par.veh.m+par.veh.Cd*v.^2+par.veh.Cr+par.veh.m*par.veh.g*gamma0;
      P=0.5*par.prb.Ts*sum(par.em.beta0*v.^2+par.em.beta1*u.*v+par.em.beta2*u.^2);
    

      %Conditions==
      if norm(aux-x) < 1e-8 && norm(P-P_old)<1e-8
          stop=1;
      end
       j=j+1;
end

 fprintf('Iterations: %d\n',j);


%% Post processiong
    

    gamma0= par.veh.beta*sin(par.veh.gamma1*s+par.veh.gamma0);
    u = a*par.veh.m+par.veh.Cd*v.^2+par.veh.Cr+par.veh.m*par.veh.g*gamma0;
    
    Power=0.5*(par.em.beta0*v.^2+par.em.beta1*u.*v+par.em.beta2*u.^2);

    
    cost =0.5* par.prb.Ts*sum(par.em.beta2*u.^2+par.em.beta1*u.*v+par.em.beta0*v.^2);
    
    fprintf('Final cost (eco-driving): %5e\n',cost);
    
    
    h = -(par.veh.beta/par.veh.gamma1)*cos(par.veh.gamma1*s+par.veh.gamma0)-(par.veh.beta/par.veh.gamma1);
    
    
    
    
%% Baseline cost (Constant speed driving)
      
 
    a_bl=zeros(N+1,1); %Due to contant sppeed
    
    v_bl=par.veh.v0;
    for i=1:1:N
        v_bl(i+1,1)=v_bl(i)+a_bl(i)*par.prb.Ts;
    end
    
    s_bl=par.veh.s0;
    for i=1:1:N
        s_bl(i+1,1)=s_bl(i)+v_bl(i)*par.prb.Ts;
    end
    
    
    gamma_bl = sin(par.veh.beta*sin(par.veh.gamma1*s_bl+par.veh.gamma0)); 
    u_bl = a_bl*par.veh.m+par.veh.Cd*v_bl.^2+par.veh.Cr+par.veh.m*par.veh.g*gamma_bl; %Applied force
    h_bl = -(par.veh.beta/par.veh.gamma1)*cos(par.veh.gamma1*s_bl+par.veh.gamma0)-(par.veh.beta/par.veh.gamma1); %Elevation profile
  
    cost_bl =0.5* par.prb.Ts*sum(par.em.beta2*u_bl.^2+par.em.beta1*u_bl.*v_bl+par.em.beta0*v_bl.^2);%Cost

    fprintf('Final cost (constant speed): %5e\n',cost_bl);
    fprintf('Energy Savings: %f2\n',(cost-cost_bl)*100/cost_bl);
    
    
    
%% Plots   
    
    
    fig = figure;
    left_color = [39, 75, 130]/255;
    right_color =  [36, 96, 63]/255;
    set(fig,'defaultAxesColorOrder',[left_color; right_color]);
    
    subplot(2,1,1)%---------------------
    
    yyaxis left
    axis([0, max(s/1000),min([v_bl;v]*3.6) ,max([v_bl;v]*3.6)])
    
    plot(s_bl/1000,v_bl*3.6,'LineWidth',4)
    hold on
    plot(s/1000,v*3.6,'-.','LineWidth',4,'Color',[67, 137, 198]/255)

    xlabel('Displacement \sl s \rm [km]')
    ylabel('Velocity \sl v \rm [km/h]')
    hold on
    
    yyaxis right
    ar=area(s_bl/1000,h_bl);
    ar.FaceAlpha =0.3;
    ar.FaceColor = [36, 96, 63]/255;
    ar.LineStyle  =':';
    grid on
    legend('Constant speed driving', 'Eco-driving');
    
    
    
   subplot(2,1,2)%------------------------------------
    
    yyaxis left
    axis([0, max(s/1000) ,min(u/1000) ,max(u/1000)])
  
    plot(s/1000,u/1000,'-.','LineWidth',4,'Color',[67, 137, 198]/255)
    xlabel('Displacement \sl s \rm [km]')
    ylabel('Traction Force \sl u \rm [kN]')
    
    yyaxis right
    ar=area(s_bl/1000,h_bl);
    ar.FaceAlpha =0.3;
    ar.FaceColor = [36, 96, 63]/255;
    ar.LineStyle  =':';
    ylabel('Elevation \sl h \rm [m]')
    grid on

    
    
    
    